# Copyright (C) 2017-present Pinterest Inc.
#
# This module is part of ptracer and is released under
# the Apache 2.0 License: http://www.apache.org/licenses/LICENSE-2.0


from ._gen_defs_linux_64 import *  # noqa


FUNCTION_REG = 'orig_rax'
ARGS_REGS = ('rdi', 'rsi', 'rdx', 'r10', 'r8', 'r9')
RETURN_REG = 'rax'
